from lxml import etree

from tests.utils import assert_nodes_equal
from zeep import xsd


def test_complex_type_alt():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <schema xmlns="http://www.w3.org/2001/XMLSchema"
                  xmlns:tns="http://tests.python-zeep.org/"
                  targetNamespace="http://tests.python-zeep.org/"
                  elementFormDefault="qualified">
            <element name="Address">
              <complexType>
                <sequence>
                  <element minOccurs="0" maxOccurs="1" name="foo" type="string" />
                </sequence>
              </complexType>
            </element>
          </schema>
        </types>
    """.strip())

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')
    obj = address_type(foo='bar')

    expected = """
      <document>
        <ns0:Address xmlns:ns0="http://tests.python-zeep.org/">
          <ns0:foo>bar</ns0:foo>
        </ns0:Address>
      </document>
    """

    node = etree.Element('document')
    address_type.render(node, obj)
    assert_nodes_equal(expected, node)


def test_element_with_annotation():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <schema xmlns="http://www.w3.org/2001/XMLSchema"
                  xmlns:tns="http://tests.python-zeep.org/"
                  targetNamespace="http://tests.python-zeep.org/"
                  elementFormDefault="qualified">
            <element name="Address" type="tns:AddressType">
                <annotation>
                    <documentation>HOI!</documentation>
                </annotation>
            </element>
            <complexType name="AddressType">
              <sequence>
                <element minOccurs="0" maxOccurs="unbounded" name="foo" type="string" />
              </sequence>
            </complexType>
          </schema>
        </types>
    """.strip())
    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')
    address_type(foo='bar')


def test_complex_type_parsexml():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <schema xmlns="http://www.w3.org/2001/XMLSchema"
                  xmlns:tns="http://tests.python-zeep.org/"
                  targetNamespace="http://tests.python-zeep.org/"
                  elementFormDefault="qualified">
            <element name="Address">
              <complexType>
                <sequence>
                  <element minOccurs="0" maxOccurs="1" name="foo" type="string" />
                </sequence>
              </complexType>
            </element>
          </schema>
        </types>
    """.strip())

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')

    input_node = etree.fromstring("""
        <Address xmlns="http://tests.python-zeep.org/">
          <foo>bar</foo>
        </Address>
    """)

    obj = address_type.parse(input_node)
    assert obj.foo == 'bar'


def test_complex_type_array_parsexml():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <schema xmlns="http://www.w3.org/2001/XMLSchema"
                  xmlns:tns="http://tests.python-zeep.org/"
                  targetNamespace="http://tests.python-zeep.org/"
                  elementFormDefault="qualified">
            <element name="Address">
              <complexType>
                <sequence>
                  <element minOccurs="0" maxOccurs="unbounded" name="foo" type="string" />
                </sequence>
              </complexType>
            </element>
          </schema>
        </types>
    """.strip())

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')

    input_node = etree.fromstring("""
        <Address xmlns="http://tests.python-zeep.org/">
          <foo>bar</foo>
          <foo>zoo</foo>
        </Address>
    """)

    obj = address_type.parse(input_node)
    assert obj.foo == ['bar', 'zoo']


def test_array():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <schema xmlns="http://www.w3.org/2001/XMLSchema"
                  xmlns:tns="http://tests.python-zeep.org/"
                  targetNamespace="http://tests.python-zeep.org/"
                  elementFormDefault="qualified">
            <element name="Address">
              <complexType>
                <sequence>
                  <element minOccurs="0" maxOccurs="unbounded" name="foo" type="string" />
                </sequence>
              </complexType>
            </element>
          </schema>
        </types>
    """.strip())

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')
    obj = address_type()
    assert obj.foo == []
    obj.foo.append('foo')
    obj.foo.append('bar')

    expected = """
        <document>
          <ns0:Address xmlns:ns0="http://tests.python-zeep.org/">
            <ns0:foo>foo</ns0:foo>
            <ns0:foo>bar</ns0:foo>
          </ns0:Address>
        </document>
    """

    node = etree.Element('document')
    address_type.render(node, obj)
    assert_nodes_equal(expected, node)


def test_array_single_init():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <schema xmlns="http://www.w3.org/2001/XMLSchema"
                  xmlns:tns="http://tests.python-zeep.org/"
                  targetNamespace="http://tests.python-zeep.org/"
                  elementFormDefault="qualified">
            <element name="Address">
              <complexType>
                <sequence>
                  <element minOccurs="0" maxOccurs="unbounded" name="foo" type="string" />
                </sequence>
              </complexType>
            </element>
          </schema>
        </types>
    """.strip())

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')
    obj = address_type(foo=['foo'])

    expected = """
        <document>
          <ns0:Address xmlns:ns0="http://tests.python-zeep.org/">
            <ns0:foo>foo</ns0:foo>
          </ns0:Address>
        </document>
    """

    node = etree.Element('document')
    address_type.render(node, obj)
    assert_nodes_equal(expected, node)


def test_array_resolve_lazy():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <schema xmlns="http://www.w3.org/2001/XMLSchema"
                  xmlns:tns="http://tests.python-zeep.org/"
                  targetNamespace="http://tests.python-zeep.org/"
                  elementFormDefault="qualified">
            <element name="Address" type="tns:AddressType" />
            <complexType name="AddressType">
              <sequence>
                <element minOccurs="0" maxOccurs="unbounded" name="foo" type="string" />
              </sequence>
            </complexType>
          </schema>
        </types>
    """.strip())

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')
    obj = address_type()
    assert obj.foo == []
    obj.foo.append('foo')
    obj.foo.append('bar')

    expected = """
        <document>
          <ns0:Address xmlns:ns0="http://tests.python-zeep.org/">
            <ns0:foo>foo</ns0:foo>
            <ns0:foo>bar</ns0:foo>
          </ns0:Address>
        </document>
    """

    node = etree.Element('document')
    address_type.render(node, obj)
    assert_nodes_equal(expected, node)


def test_complex_type_array_to_other_complex_object():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
            <xs:complexType name="Address">
              <xs:sequence>
                <xs:element minOccurs="0" maxOccurs="1" name="foo" type="xs:string" />
              </xs:sequence>
            </xs:complexType>
            <xs:complexType name="ArrayOfAddress">
              <xs:sequence>
                <xs:element minOccurs="0" maxOccurs="unbounded" name="Address" nillable="true" type="Address" />
              </xs:sequence>
            </xs:complexType>
            <xs:element name="ArrayOfAddress" type="ArrayOfAddress"/>
          </xs:schema>
        </types>
    """.strip())  # noqa

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_array = schema.get_element('ArrayOfAddress')
    obj = address_array()
    assert obj.Address == []

    obj.Address.append(schema.get_type('Address')(foo='foo'))
    obj.Address.append(schema.get_type('Address')(foo='bar'))

    node = etree.fromstring("""
        <?xml version="1.0"?>
        <ArrayOfAddress>
            <Address>
                <foo>foo</foo>
            </Address>
            <Address>
                <foo>bar</foo>
            </Address>
        </ArrayOfAddress>
    """.strip())


def test_complex_type_init_kwargs():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <schema xmlns="http://www.w3.org/2001/XMLSchema"
                  xmlns:tns="http://tests.python-zeep.org/"
                  targetNamespace="http://tests.python-zeep.org/">
            <element name="Address">
              <complexType>
                <sequence>
                  <element minOccurs="0" maxOccurs="1" name="NameFirst" type="string"/>
                  <element minOccurs="0" maxOccurs="1" name="NameLast" type="string"/>
                  <element minOccurs="0" maxOccurs="1" name="Email" type="string"/>
                </sequence>
              </complexType>
            </element>
          </schema>
        </types>
    """.strip())

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')
    obj = address_type(
        NameFirst='John', NameLast='Doe', Email='j.doe@example.com')
    assert obj.NameFirst == 'John'
    assert obj.NameLast == 'Doe'
    assert obj.Email == 'j.doe@example.com'


def test_complex_type_init_args():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <schema xmlns="http://www.w3.org/2001/XMLSchema"
                  xmlns:tns="http://tests.python-zeep.org/"
                  targetNamespace="http://tests.python-zeep.org/">
            <element name="Address">
              <complexType>
                <sequence>
                  <element minOccurs="0" maxOccurs="1" name="NameFirst" type="string"/>
                  <element minOccurs="0" maxOccurs="1" name="NameLast" type="string"/>
                  <element minOccurs="0" maxOccurs="1" name="Email" type="string"/>
                </sequence>
              </complexType>
            </element>
          </schema>
        </types>
    """.strip())

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')
    obj = address_type('John', 'Doe', 'j.doe@example.com')
    assert obj.NameFirst == 'John'
    assert obj.NameLast == 'Doe'
    assert obj.Email == 'j.doe@example.com'


def test_complex_type_with_attributes():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
            <xs:complexType name="Address">
              <xs:sequence>
                <xs:element minOccurs="0" maxOccurs="1" name="NameFirst" type="xs:string"/>
                <xs:element minOccurs="0" maxOccurs="1" name="NameLast" type="xs:string"/>
                <xs:element minOccurs="0" maxOccurs="1" name="Email" type="xs:string"/>
              </xs:sequence>
              <xs:attribute name="id" type="xs:string" use="required"/>
            </xs:complexType>
            <xs:element name="Address" type="Address"/>
          </xs:schema>
        </types>
    """.strip())

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))

    address_type = schema.get_element('Address')
    obj = address_type(
        NameFirst='John', NameLast='Doe', Email='j.doe@example.com', id='123')

    node = etree.Element('document')
    address_type.render(node, obj)

    expected = """
        <document>
            <Address id="123">
                <NameFirst>John</NameFirst>
                <NameLast>Doe</NameLast>
                <Email>j.doe@example.com</Email>
            </Address>
        </document>
    """
    assert_nodes_equal(expected, node)


def test_complex_type_with_extension():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <xs:schema
              xmlns:xs="http://www.w3.org/2001/XMLSchema"
              xmlns:tns="http://tests.python-zeep.org/"
              targetNamespace="http://tests.python-zeep.org/"
              elementFormDefault="qualified">

              <xs:complexType name="Address">
                <xs:complexContent>
                  <xs:extension base="tns:Name">
                    <xs:sequence>
                      <xs:element name="country" type="xs:string"/>
                    </xs:sequence>
                  </xs:extension>
                </xs:complexContent>
              </xs:complexType>
            <xs:element name="Address" type="tns:Address"/>

            <xs:complexType name="Name">
              <xs:sequence>
                <xs:element name="first_name" type="xs:string"/>
                <xs:element name="last_name" type="xs:string"/>
              </xs:sequence>
            </xs:complexType>
          </xs:schema>
        </types>
    """.strip())
    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')

    obj = address_type(
        first_name='foo', last_name='bar', country='The Netherlands')

    node = etree.Element('document')
    address_type.render(node, obj)
    expected = """
        <document>
            <ns0:Address xmlns:ns0="http://tests.python-zeep.org/">
                <ns0:first_name>foo</ns0:first_name>
                <ns0:last_name>bar</ns0:last_name>
                <ns0:country>The Netherlands</ns0:country>
            </ns0:Address>
        </document>
    """
    assert_nodes_equal(expected, node)


def test_complex_type_simple_content():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <xs:schema
              xmlns:xs="http://www.w3.org/2001/XMLSchema"
              xmlns:tns="http://tests.python-zeep.org/"
              targetNamespace="http://tests.python-zeep.org/"
              elementFormDefault="qualified">

            <xs:element name="ShoeSize">
              <xs:complexType>
                <xs:simpleContent>
                  <xs:extension base="xs:integer">
                    <xs:attribute name="sizing" type="xs:string" />
                  </xs:extension>
                </xs:simpleContent>
              </xs:complexType>
            </xs:element>
          </xs:schema>
        </types>
    """.strip())
    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    shoe_type = schema.get_element('{http://tests.python-zeep.org/}ShoeSize')

    obj = shoe_type(20, sizing='EUR')

    node = etree.Element('document')
    shoe_type.render(node, obj)
    expected = """
        <document>
            <ns0:ShoeSize xmlns:ns0="http://tests.python-zeep.org/" sizing="EUR">20</ns0:ShoeSize>
        </document>
    """
    assert_nodes_equal(expected, node)


def test_custom_simple_type():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <schema xmlns="http://www.w3.org/2001/XMLSchema"
                  xmlns:tns="http://tests.python-zeep.org/"
                  targetNamespace="http://tests.python-zeep.org/"
                  elementFormDefault="qualified">
            <element name="something">
              <simpleType>
                <restriction base="integer">
                  <minInclusive value="0"/>
                  <maxInclusive value="100"/>
                </restriction>
              </simpleType>
            </element>
          </schema>
        </types>
    """.strip())

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))

    custom_type = schema.get_element('{http://tests.python-zeep.org/}something')
    obj = custom_type(75)

    node = etree.Element('document')
    custom_type.render(node, obj)
    expected = """
        <document>
            <ns0:something xmlns:ns0="http://tests.python-zeep.org/">75</ns0:something>
        </document>
    """
    assert_nodes_equal(expected, node)


def test_group():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema"
                     xmlns:tns="http://tests.python-zeep.org/"
                     targetNamespace="http://tests.python-zeep.org/"
                     elementFormDefault="qualified">

            <xs:element name="Address">
              <xs:complexType>
                <xs:group ref="tns:Name" />
              </xs:complexType>
            </xs:element>

            <xs:group name="Name">
              <xs:sequence>
                <xs:element name="first_name" type="xs:string" />
                <xs:element name="last_name" type="xs:string" />
              </xs:sequence>
            </xs:group>

          </xs:schema>
        </types>
    """.strip())
    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')

    obj = address_type(first_name='foo', last_name='bar')

    node = etree.Element('document')
    address_type.render(node, obj)
    expected = """
        <document>
            <ns0:Address xmlns:ns0="http://tests.python-zeep.org/">
                <ns0:first_name>foo</ns0:first_name>
                <ns0:last_name>bar</ns0:last_name>
            </ns0:Address>
        </document>
    """
    assert_nodes_equal(expected, node)


def test_group_for_type():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema"
                     xmlns:tns="http://tests.python-zeep.org/"
                     targetNamespace="http://tests.python-zeep.org/"
                     elementFormDefault="unqualified">

            <xs:element name="Address" type="tns:AddressType" />

            <xs:complexType name="AddressType">
              <xs:sequence>
                <xs:group ref="tns:NameGroup"/>
                <xs:group ref="tns:AddressGroup"/>
              </xs:sequence>
            </xs:complexType>

            <xs:group name="NameGroup">
              <xs:sequence>
                <xs:element name="first_name" type="xs:string" />
                <xs:element name="last_name" type="xs:string" />
              </xs:sequence>
            </xs:group>

            <xs:group name="AddressGroup">
              <xs:annotation>
                <xs:documentation>blub</xs:documentation>
              </xs:annotation>
              <xs:sequence>
                <xs:element name="city" type="xs:string" />
                <xs:element name="country" type="xs:string" />
              </xs:sequence>
            </xs:group>
          </xs:schema>
        </types>
    """.strip())
    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')

    obj = address_type(
        first_name='foo', last_name='bar',
        city='Utrecht', country='The Netherlands')

    node = etree.Element('document')
    address_type.render(node, obj)
    expected = """
        <document>
            <ns0:Address xmlns:ns0="http://tests.python-zeep.org/">
                <first_name>foo</first_name>
                <last_name>bar</last_name>
                <city>Utrecht</city>
                <country>The Netherlands</country>
            </ns0:Address>
        </document>
    """
    assert_nodes_equal(expected, node)


def test_element_ref():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <schema xmlns="http://www.w3.org/2001/XMLSchema"
                  xmlns:tns="http://tests.python-zeep.org/"
                  targetNamespace="http://tests.python-zeep.org/"
                     elementFormDefault="qualified">
            <element name="foo" type="string"/>
            <element name="bar">
              <complexType>
                <sequence>
                  <element ref="tns:foo"/>
                </sequence>
              </complexType>
            </element>
          </schema>
        </types>
    """.strip())

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))

    foo_type = schema.get_element('{http://tests.python-zeep.org/}foo')
    assert isinstance(foo_type.type, xsd.String)

    custom_type = schema.get_element('{http://tests.python-zeep.org/}bar')
    obj = custom_type(foo='bar')

    node = etree.Element('document')
    custom_type.render(node, obj)
    expected = """
        <document>
            <ns0:bar xmlns:ns0="http://tests.python-zeep.org/">
                <ns0:foo>bar</ns0:foo>
            </ns0:bar>
        </document>
    """
    assert_nodes_equal(expected, node)


def test_unqualified():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <types>
          <schema xmlns="http://www.w3.org/2001/XMLSchema"
                  xmlns:xsd="http://www.w3.org/2001/XMLSchema"
                  xmlns:tns="http://tests.python-zeep.org/"
                  attributeFormDefault="qualified"
                  elementFormDefault="qualified"
                  targetNamespace="http://tests.python-zeep.org/">
            <element name="Address">
              <complexType>
                <sequence>
                  <element name="foo" type="xsd:string" form="unqualified" />
                </sequence>
              </complexType>
            </element>
          </schema>
        </types>
    """.strip())

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')
    obj = address_type(foo='bar')

    expected = """
      <document>
        <ns0:Address xmlns:ns0="http://tests.python-zeep.org/">
          <foo>bar</foo>
        </ns0:Address>
      </document>
    """

    node = etree.Element('document')
    address_type.render(node, obj)
    assert_nodes_equal(expected, node)


def test_qualified_attribute():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <wsdl:types xmlns:wsdl="http://schemas.xmlsoap.org/wsdl/">
          <xsd:schema
                  xmlns:xsd="http://www.w3.org/2001/XMLSchema"
                  xmlns:tns="http://tests.python-zeep.org/"
                  attributeFormDefault="qualified"
                  elementFormDefault="qualified"
                  targetNamespace="http://tests.python-zeep.org/">
            <xsd:element name="Address">
              <xsd:complexType>
                <xsd:sequence>
                  <xsd:element name="foo" type="xsd:string" form="unqualified" />
                </xsd:sequence>
                <xsd:attribute name="id" type="xsd:string" use="required" form="unqualified" />
                <xsd:attribute name="pos" type="xsd:string" use="required" />
              </xsd:complexType>
            </xsd:element>
          </xsd:schema>
        </wsdl:types>
    """.strip())

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')
    obj = address_type(foo='bar', id="20", pos="30")

    expected = """
      <document>
        <ns0:Address xmlns:ns0="http://tests.python-zeep.org/" id="20" ns0:pos="30">
          <foo>bar</foo>
        </ns0:Address>
      </document>
    """

    node = etree.Element('document')
    address_type.render(node, obj)
    assert_nodes_equal(expected, node)


def test_init_with_dicts():
    node = etree.fromstring("""
        <?xml version="1.0"?>
        <wsdl:types xmlns:wsdl="http://schemas.xmlsoap.org/wsdl/">
          <xsd:schema
                  xmlns:xsd="http://www.w3.org/2001/XMLSchema"
                  xmlns:tns="http://tests.python-zeep.org/"
                  attributeFormDefault="qualified"
                  elementFormDefault="qualified"
                  targetNamespace="http://tests.python-zeep.org/">
            <xsd:element name="Address">
              <xsd:complexType>
                <xsd:sequence>
                  <xsd:element name="name" type="xsd:string"/>
                  <xsd:element minOccurs="0" name="optional" type="xsd:string"/>
                  <xsd:element name="container" nillable="true" type="tns:Container"/>
                </xsd:sequence>
              </xsd:complexType>
            </xsd:element>

            <xsd:complexType name="Container">
              <xsd:sequence>
                <xsd:element maxOccurs="unbounded" minOccurs="0" name="service"
                             nillable="true" type="tns:ServiceRequestType"/>
              </xsd:sequence>
            </xsd:complexType>

            <xsd:complexType name="ServiceRequestType">
              <xsd:sequence>
                <xsd:element name="name" type="xsd:string"/>
              </xsd:sequence>
            </xsd:complexType>
          </xsd:schema>
        </wsdl:types>
    """.strip())

    schema = xsd.Schema(node.find('{http://www.w3.org/2001/XMLSchema}schema'))
    address_type = schema.get_element('{http://tests.python-zeep.org/}Address')
    obj = address_type(name='foo', container={'service': [{'name': 'foo'}]})

    expected = """
      <document>
        <ns0:Address xmlns:ns0="http://tests.python-zeep.org/">
          <ns0:name>foo</ns0:name>
          <ns0:container>
            <ns0:service>
              <ns0:name>foo</ns0:name>
            </ns0:service>
          </ns0:container>
        </ns0:Address>
      </document>
    """

    node = etree.Element('document')
    address_type.render(node, obj)
    assert_nodes_equal(expected, node)
